#   -*- coding: utf-8 -*-
#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2020-2021 John Mille <john@compose-x.io>


"""
Module with functions that can be used as AWS Lambda Handlers to perform various tasks
"""

import logging
import sys
from os import environ
from time import sleep

from jsonschema import validate

from .kafka_connect_api import Api, Cluster, Connector

KEYISSET = lambda x, y: isinstance(y, dict) and x in y.keys() and y[x]

CLUSTER_CONFIG_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$id": "kafka-connect-cluster-lambda-config",
    "type": "object",
    "title": "kafka-connect-cluster-lambda-config",
    "description": "Schema of the event payload expected to interact with the Lambda Function",
    "required": ["hostname"],
    "properties": {
        "hostname": {"type": "string", "format": "idn-hostname"},
        "port": {"type": "number", "default": 8083},
        "url": {
            "type": "string",
            "description": "The URL to use. Overrides hostname:port",
        },
        "username": {"type": "string"},
        "password": {"type": "string"},
        "protocol": {"type": "string", "default": "http"},
        "ignore_ssl_errors": {"type": "boolean", "default": False},
    },
}


CONNECTOR_CONFIG_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",
    "$id": "kafka-connect-connector-lambda-config",
    "type": "object",
    "title": "kafka-connect-connector-lambda-config",
    "description": "Schema of the event configuration required for a connector",
    "required": ["name"],
    "properties": {
        "name": {
            "type": "string",
        },
        "config": {"type": "object"},
    },
}


def setup_logging():
    """
    In case this is used in a Lambda function, removes the AWS Lambda default log handler

    :returns: the_logger
    :rtype: Logger
    """
    formats = {
        "INFO": logging.Formatter(
            "[%(levelname)8s] %(message)s",
            "%Y-%m-%d %H:%M:%S",
        )
    }
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)
    formatter = formats["INFO"]
    for handler in root_logger.handlers:
        handler.setFormatter(formatter)
        root_logger.addHandler(handler)
    return root_logger


def config_from_env_vars():
    """
    Sets the config from environment variables
    :return: the cluster configuration
    :rtype: dict
    """
    if (
        environ.get("CONNECT_CLUSTER_HOSTNAME", None) is None
        and environ.get("CONNECT_CLUSTER_URL", None) is None
    ):
        raise LookupError("CONNECT_CLUSTER_HOSTNAME nor CONNECT_CLUSTER_URL are set.")
    return {
        "hostname": environ.get("CONNECT_CLUSTER_HOSTNAME", None),
        "port": int(environ.get("CONNECT_CLUSTER_PORT", 8083)),
        "url": environ.get("CONNECT_CLUSTER_URL", None),
        "username": environ.get("CONNECT_CLUSTER_BASIC_AUTH_USERNAME", None),
        "password": environ.get("CONNECT_CLUSTER_BASIC_AUTH_PASSWORD", None),
        "protocol": environ.get("CONNECT_CLUSTER_PROTOCOL", "http"),
        "ignore_ssl_errors": environ.get("CONNECT_CLUSTER_IGNORE_SSL_ERRORS", False),
    }


def config_from_event(event):
    """
    Validates the cluster configuration against schema if passed in event

    :param dict event:
    :return: The connect cluster configuration
    :rtype: dict
    """
    if not KEYISSET("cluster", event):
        return None
    cluster_config = event["cluster"]
    validate(cluster_config, CLUSTER_CONFIG_SCHEMA)
    return cluster_config


def set_cluster_config(event):
    """
    Function to set the connect cluster settings

    :param dict event:
    :return: The cluster config
    :rtype: dict
    """
    cluster_config = config_from_event(event)
    if not cluster_config:
        cluster_config = config_from_env_vars()
    if not cluster_config:
        raise ValueError("No connect cluster configuration was given.")
    return cluster_config


def get_connector(event):
    """
    Function to retrieve the one connector in the connect cluster from event

    :param dict event:
    :return: the connector
    :rtype: Connector
    """
    log = setup_logging()
    connector_config = event["connector"]
    validate(connector_config, CONNECTOR_CONFIG_SCHEMA)
    name = connector_config["name"]
    cluster_config = set_cluster_config(event)
    api = Api(**cluster_config)
    cluster = Cluster(api)
    log.info(cluster)
    if name not in cluster.connectors:
        raise KeyError(f"Connector {name} is not present in cluster {cluster}")
    return cluster.connectors[name]


def restart_all_connectors(event, context):
    """
    Function to restart all the connectors in a Connect cluster

    :param dict event:
    :param dict context:
    """
    log = setup_logging()
    cluster_config = set_cluster_config(event)
    api = Api(**cluster_config)
    cluster = Cluster(api)
    log.info(f"Restarting all connectors in {cluster}")
    log.info(cluster)
    for connector in cluster.connectors.values():
        log.info(f"Restarting {connector}")
        connector.cycle_connector()
    return 0


def create_update_connector(event, context):
    """
    Function to create / update a new connector

    :param dict event:
    :param dict context:
    """
    log = setup_logging()
    if not KEYISSET("connector", event):
        raise KeyError("No configuration given for the connector to set/update")
    connector_config = event["connector"]
    validate(connector_config, CONNECTOR_CONFIG_SCHEMA)
    name = connector_config["name"]
    if not KEYISSET("config", connector_config):
        raise KeyError(
            f"The connector {name} has no configuration passed in event for create/update"
        )
    cluster_config = set_cluster_config(event)
    api = Api(**cluster_config)
    cluster = Cluster(api)
    log.info(f"Attempt at creating/updating {name} in {cluster}")
    connector = Connector(api, name)
    connector.config = connector_config["config"]
    log.info(f"{connector.name} - waiting 5s before checking")
    sleep(5)
    if name not in cluster.connectors:
        log.error(f"Failed to create connector {name} in {cluster}")
        raise Exception(f"Failed to create connector {name}")
    return 0


def delete_connector(event, context):
    """
    Function to delete a connector. No need for config

    :param dict event:
    :param dict context:
    """
    log = setup_logging()
    connector = get_connector(event)
    log.info(f"Deleting connector {connector}")
    connector.delete()
    return 0


def restart_connector(event, context):
    """
    Function to delete a connector. No need for config

    :param dict event:
    :param dict context:
    """
    log = setup_logging()
    connector = get_connector(event)
    log.info(
        f"Cycling connector {connector} through pause / restart all tasks / resume"
    )
    connector.cycle_connector()
    return 0


def check_connector_health(event, context):
    """
    Function to evaluate the connector health

    :param dict event:
    :param dict context:
    :return:
    """
    log = setup_logging()
    tasks_health = []
    connector = get_connector(event)
    for task in connector.tasks:
        log.info(f"Task {task.id} for connector {connector} state is {task.state}")
        tasks_health.append(task.is_running())
    if all(tasks_health):
        log.info(f"All tasks for connector {connector} are RUNNING")
        return True
    return False



#  -*- coding: utf-8 -*-
# SPDX-License-Identifier: MPL-2.0
# Copyright 2020-2021 John Mille <john@compose-x.io>

"""Main module."""
import re

import requests
from requests.auth import HTTPBasicAuth


class Task(object):
    """
    Class to represent a Connector Task
    """

    def __init__(self, api, connector, task_id, task_config):
        """

        :param Api api:
        :param Connector connector:
        :param int task_id:
        """
        self.id = task_id
        self._connector = connector
        self._api = api
        self.config = task_config

    def __repr__(self):
        return f"{self._connector.name}.{self.id}"

    @property
    def status(self):
        _query = self._api.get(
            f"/connectors/{self._connector.name}/tasks/{self.id}/status"
        )
        return _query

    @property
    def state(self):
        return self.status["state"]

    def is_running(self):
        if self.state == "RUNNING":
            return True
        return False

    def restart(self):
        _query = self._api.post_raw(
            f"/connectors/{self._connector.name}/tasks/{self.id}/restart"
        )


class Connector(object):
    """
    Class to represent a Connector
    """

    def __init__(self, api, name, config=None):
        """

        :param Api api:
        :param str name:
        :param dict config:
        """
        self._api = api
        self.name = name

    def __repr__(self):
        return self.name

    def exists(self):
        req = self._api.get_raw(f"/connectors/{self.name}/")
        if req.status_code == 404:
            return False
        return True

    def restart(self):
        self._api.post(f"/connectors/{self.name}/restart")

    def pause(self):
        self._api.put_raw(f"/connectors/{self.name}/pause")

    def resume(self):
        self._api.put_raw(f"/connectors/{self.name}/resume")

    def restart_all_tasks(self):
        for _task in self.tasks:
            _task.restart()

    def delete(self):
        self._api.delete_raw(f"/connectors/{self.name}")

    def cycle_connector(self):
        self.pause()
        self.restart_all_tasks()
        self.resume()

    @property
    def state(self):
        return self.status()["connector"]["state"]

    @property
    def config(self):
        _config = self._api.get(f"/connectors/{self.name}")
        return _config["config"]

    @config.setter
    def config(self, config):
        _req = self._api.put_raw(f"/connectors/{self.name}/config", json=config)
        if not (199 < _req.status_code < 300):
            print(_req.text)

    @property
    def status(self):
        return self._api.get(f"/connectors/{self.name}/status")

    @property
    def tasks(self):
        _tasks = []
        _connector_tasks = self._api.get(f"/connectors/{self.name}/tasks")
        for _task in _connector_tasks:
            _tasks.append(
                Task(
                    self._api,
                    self,
                    task_id=int(_task["id"]["task"]),
                    task_config=_task["config"],
                )
            )
        return _tasks


class Cluster(object):
    def __init__(self, api):
        self._api = api

    def get(self):
        return self._api.get("/")

    @property
    def version(self):
        return self.get()["version"]

    @property
    def kafka_cluster(self):
        return self.get()["kafka_cluster_id"]

    @property
    def connectors(self):
        _connectors = self._api.get("/connectors")
        _cluster_connectors = {}
        for connector in _connectors:
            _cluster_connectors[connector] = Connector(self._api, connector)
        return _cluster_connectors

    def __repr__(self):
        return f"{self._api.url} - {self.version} - {self.kafka_cluster}"


class Api(object):
    """
    Class to represent the Connect cluster
    """

    def __init__(
        self,
        hostname,
        port=None,
        url=None,
        protocol=None,
        ignore_ssl_errors=False,
        username=None,
        password=None,
    ):
        """

        :param str hostname:
        :param int port:
        :param str protocol:
        :param bool ignore_ssl_errors:
        :param str username:
        :param str password:
        """
        self.hostname = hostname
        self.protocol = protocol if protocol else "http"
        self.verify_ssl = not ignore_ssl_errors
        self.port = port if port else 8083
        self.username = username
        self.password = password

        if self.protocol not in ["http", "https"]:
            raise ValueError("protocol must be one of", ["http", "https"])
        if (self.port < 0) or (self.port > (2 ** 16)):
            raise ValueError(
                f"Port {self.port} is not valid. Must be between 0 and {((2 ** 16) - 1)}"
            )
        if self.username and not self.password or self.password and not self.username:
            raise ValueError("You must specify both username and password")
        if self.verify_ssl is True and self.protocol == "http":
            print("No SSL needed for HTTP without TLS. Disabling")
            self.verify_ssl = False
        if url:
            self.url = url
            if not re.match(r"(http://|https://)", self.url):
                print(f"URL Does not contain a protocol. Using default {self.protocol}")
                self.url = f"{self.protocol}://{self.url}"
            print("URL Defined from parameter. Skipping hostname:port parameters")
        elif (self.port == 80 and protocol == "http") or (
            self.port == 443 and self.protocol == "https"
        ):
            self.url = f"{self.protocol}://{self.hostname}"
        else:
            self.url = f"{self.protocol}://{self.hostname}:{self.port}"

        self.auth = (
            HTTPBasicAuth(self.username, self.password)
            if self.username and self.password
            else None
        )

        self.headers = {
            "Content-type": "application/json",
            "Accept": "application/json",
        }

    def __repr__(self):
        return self.url

    def get_raw(self, query_path, **kwargs):
        if not query_path.startswith(r"/"):
            query_path = f"/{query_path}"
        url = f"{self.url}{query_path}"
        req = requests.get(
            url, auth=self.auth, headers=self.headers, verify=self.verify_ssl, **kwargs
        )
        return req

    def get(self, query_path):
        req = self.get_raw(query_path)
        return req.json()

    def post_raw(self, query_path, **kwargs):
        if not query_path.startswith(r"/"):
            query_path = f"/{query_path}"
        url = f"{self.url}{query_path}"
        req = requests.post(
            url,
            auth=self.auth,
            headers=self.headers,
            verify=self.verify_ssl,
            **kwargs,
        )
        return req

    def post(self, query_path, **kwargs):
        req = self.post_raw(query_path, **kwargs)
        return req.json()

    def put_raw(self, query_path, **kwargs):
        if not query_path.startswith(r"/"):
            query_path = f"/{query_path}"
        url = f"{self.url}{query_path}"
        req = requests.put(
            url,
            auth=self.auth,
            headers=self.headers,
            verify=self.verify_ssl,
            **kwargs,
        )
        return req

    def put(self, query_path, **kwargs):
        req = self.put_raw(query_path, **kwargs)
        return req.json()

    def delete_raw(self, query_path, **kwargs):
        if not query_path.startswith(r"/"):
            query_path = f"/{query_path}"
        url = f"{self.url}{query_path}"
        req = requests.delete(
            url, auth=self.auth, headers=self.headers, verify=self.verify_ssl, **kwargs
        )
        return req

    def delete(self, query_path):
        req = self.delete_raw(query_path)
        return req.json()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['kafka_connect_api']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.26.0,<3.0.0']

extras_require = \
{':extra == "awslambda"': ['jsonschema>=4.3.0,<5.0.0']}

setup_kwargs = {
    'name': 'kafka-connect-api',
    'version': '0.2.0',
    'description': 'Python Client to interact with Apache Kafka Connect cluster',
    'long_description': None,
    'author': 'John Preston',
    'author_email': 'john@compose-x.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
